#include "IZ_net.h"

void IZ_NetLoadConfig(IZ_NetState* state, const char* config_path) {
	char buffer[32];

	ini_gets("Network", "Username", IZ_NET_DEFAULT_STATE.config.username, buffer, 32, config_path);
	memcpy_s(state->config.username, 32, buffer, 32);

	state->config.packet_interval_ms = ini_getl("Network", "PacketIntervalMs", IZ_NET_DEFAULT_STATE.config.packet_interval_ms, config_path);
	if (!(100 <= state->config.packet_interval_ms && state->config.packet_interval_ms <= 500)) {
		state->config.packet_interval_ms = IZ_NET_DEFAULT_STATE.config.packet_interval_ms;
	}

	state->config.max_reconnect_retries = ini_getl("Network", "MaxReconnectRetries", IZ_NET_DEFAULT_STATE.config.max_reconnect_retries, config_path);
	if (!(0 <= state->config.max_reconnect_retries && state->config.max_reconnect_retries <= 8)) {
		state->config.max_reconnect_retries = IZ_NET_DEFAULT_STATE.config.max_reconnect_retries;
	}

	state->config.reconnect_interval_secs = ini_getl("Network", "ReconnectIntervalSeconds", IZ_NET_DEFAULT_STATE.config.reconnect_interval_secs, config_path);
	if (!(3 <= state->config.reconnect_interval_secs && state->config.reconnect_interval_secs <= 10)) {
		state->config.reconnect_interval_secs = IZ_NET_DEFAULT_STATE.config.reconnect_interval_secs;
	}
}

IZ_ProcedureResult IZ_NetSaveConfig(IZ_NetState* state, const char* config_path) {
	if (!ini_puts("Network", "Username", state->config.username, config_path)) {
		return -1;
	}

	if (!ini_putl("Network", "PacketIntervalMs", state->config.packet_interval_ms, config_path)) {
		return -1;
	}

	if (!ini_putl("Network", "MaxReconnectRetries", state->config.max_reconnect_retries, config_path)) {
		return -1;
	}

	if (!ini_putl("Network", "ReconnectIntervalSeconds", state->config.reconnect_interval_secs, config_path)) {
		return -1;
	}

	return 0;
}

void IZ_NetOverrideConfig(IZ_NetState* state, u8 argc, const char* argv[]) {
	const char* cmdline_buffer;
	if ((cmdline_buffer = IZ_ConfigGetCommandlineOption(argc, argv, "-i"))) {
		state->config.packet_interval_ms = atoi(cmdline_buffer);
	}
}

IZ_ProcedureResult IZ_NetInitialize(
	IZ_NetState* state,
	void* user_data,
	void* callback,
	const char* config_path,
	u8 argc,
	const char* argv[]
) {
	memcpy_s(state, sizeof(IZ_NetState), &IZ_NET_DEFAULT_STATE, sizeof(IZ_NetState));
	IZ_NetLoadConfig(state, config_path);
	if (IZ_NetSaveConfig(state, config_path) < 0) {
		return -1;
	}
	IZ_NetOverrideConfig(state, argc, argv);
	if (!user_data) {
		return -2;
	}
	state->binding.user_data = user_data;
	state->callback = callback;
	state->retries = state->config.max_reconnect_retries;
	u8 player_index;
	for (player_index = 0; player_index < IZ_PLAYERS; player_index += 1) {
		state->action[player_index] = 0;
	}

	return 0;
}

void IZ_NetConnect(IZ_NetState* state, IZ_WSClientInitializeParams params) {
	if (!state->callback) {
		return;
	}

	if (state->status == IZ_NET_STATUS_CONNECTED) {
		return;
	}

	if (state->status == IZ_NET_STATUS_CONNECTING) {
		return;
	}

	if (state->retries < state->config.max_reconnect_retries) {
		return;
	}

	state->retries = 0;
	state->params = params;
	state->client_thread = SDL_CreateThread(state->callback, "networking", state->binding.user_data);
	SDL_DetachThread(state->client_thread);
}

void IZ_NetDisconnect(IZ_NetState* state) {
	if (state->status == IZ_NET_STATUS_PRISTINE) {
		return;
	}
	if (state->binding.connection) {
		IZ_WSClientVHostData *vhd = (IZ_WSClientVHostData *) lws_protocol_vh_priv_get(
			lws_get_vhost(state->binding.connection),
			lws_get_protocol(state->binding.connection)
		);
		if (vhd) {
			lws_sul_cancel(&vhd->sul);
		}
	}
	state->retries = state->config.max_reconnect_retries;
	IZ_WSClientCancelService(&state->binding);
}

void IZ_NetSendBinaryMessage(IZ_NetState* state, void* in, size_t len) {
	if (state->status != IZ_NET_STATUS_CONNECTED) {
		return;
	}

	IZ_WSClientVHostData* vhd = (IZ_WSClientVHostData*) lws_protocol_vh_priv_get(
		lws_get_vhost(state->binding.connection),
		lws_get_protocol(state->binding.connection)
	);

	IZ_WebsocketMessage amsg;
	i32 result;
	result = IZ_WebsocketCreateBinaryMessage(state->binding.connection, &amsg, in, len);

	if (result < 0) {
		return;
	}

	lws_ring_insert(vhd->ring, &amsg, 1);
	lws_callback_on_writable(state->binding.connection);
}

void IZ_NetSendTextMessage(IZ_NetState* state, char* in, size_t len) {
	if (state->status != IZ_NET_STATUS_CONNECTED) {
		return;
	}

	IZ_WSClientVHostData* vhd = (IZ_WSClientVHostData*) lws_protocol_vh_priv_get(
		lws_get_vhost(state->binding.connection),
		lws_get_protocol(state->binding.connection)
	);

	IZ_WebsocketMessage amsg;
	i32 result;
	result = IZ_WebsocketCreateTextMessage(state->binding.connection, &amsg, in, len);

	if (result < 0) {
		return;
	}

	lws_ring_insert(vhd->ring, &amsg, 1);
	lws_callback_on_writable(state->binding.connection);
}
